#ifndef __MD_LIB_VARIANT_H__
#define __MD_LIB_VARIANT_H__

#include <cassert>
#include <cstdlib>
#include <iostream>
#include <map>
#include <stdexcept>
#include <string>
#include <typeinfo>
#include <MD/private/prefix.h>
#include <MD/MD_sys_types.h>
#include <MD/MD_DateTime.h>
#include <MD/MD_EpochNS.h>
#include <MD/MD_Fracker.h>

namespace MD {
  
  // Thrown when the Variant class encounters a type that it is not
  // equiped to handle.
  class DynamicTypeError : public std::logic_error {
  public:
    DynamicTypeError() throw() : logic_error("Dynamic Typing Error") { }
    DynamicTypeError(const DynamicTypeError&) throw()
      : logic_error("Dynamic Typing Error") { }
    DynamicTypeError& operator=( const DynamicTypeError&) throw()
      { return *this; } 
    virtual ~DynamicTypeError() throw() { }
  };
  
  // Thrown when there is an improper attempt to cast from types that
  // Variant is equiped to handle such as a map to an int
  class BadVariantCast : public std::bad_cast {
  public:
    virtual const char * what() const throw()
      { return "Dynamic Typing Error: bad variant cast"; }
  };

  // Warning: using this class can result in a LOT of memory copying.
  // Pass by reference.  a simple struct that just has data and its
  // size
  struct BufferHolder {
    BufferHolder() : size(0), data(NULL) { }
    BufferHolder(size_t s, const void *d);
    BufferHolder(const BufferHolder& buff);
    
    BufferHolder& operator=(const BufferHolder& copy);
    ~BufferHolder() {
      size = 0;
      std::free(data);
    }
    unsigned int size;
    void *data;
  };
  
  // enum of the types supported by Variant
  enum VarType {
    DUMMY = -1,
    CHAR = 1,
    INT = 2,
    SHORT = 3,
    FLOAT = 4,
    DOUBLE = 5,
    STRING = 6,
    MAP = 13,
    BUFFER = 14,
    UINT = 16,
    UINT64 = 18,
    INT64 = 19,
    DATETIME = 20,
    EPOCH_NSEC = 22
  };

  class MD_Variant {
  public:
    MD_Variant();
    ~MD_Variant();
    MD_Variant(const MD_Variant& copy);

    MD_Variant& operator=(const MD_Variant& source);
    MD_Variant& operator=(char c);
    MD_Variant& operator=(short s);
    MD_Variant& operator=(int i);
    MD_Variant& operator=(int64_t i);
    MD_Variant& operator=(float f);
    MD_Variant& operator=(double d);
    MD_Variant& operator=(MD_DateTime dt);
    MD_Variant& operator=(const MD_EpochNS &dt);
    MD_Variant& operator=(const std::string& str);
    MD_Variant& operator=(const char * str);
    MD_Variant& operator=(const BufferHolder& buff);
    MD_Variant& operator=(const std::map<MD_Variant, MD_Variant>& map);
    MD_Variant& operator=(unsigned int u);
    MD_Variant& operator=(uint64_t u);

    // Old-style Holder interface:
#ifndef NO_HOLDER_COMPATABILITY
    MD_Variant(char c);
    MD_Variant(unsigned int i);
    MD_Variant(uint64_t i);
    MD_Variant(float f);
    MD_Variant(double d);
    MD_Variant(int i);
    MD_Variant(int64_t i);
    MD_Variant(MD_EpochNS ent);
    MD_Variant(MD_DateTime dt);
    MD_Variant(const std::string& s);
    MD_Variant(const char *s);

    int copy(const unsigned char* s, int typ, int len);
    int copy(MD_Variant& source)
      { operator=(source); return true; }

    const void * getDataPtr() const;
    size_t len() const;

    operator char() const
      { return Char(); }

    operator int() const
      { return Int(); }

    operator unsigned int() const
      { return UnsignedInt(); }

    operator int64_t() const
      { return Int64(); }

    operator uint64_t() const
      { return UnsignedInt64(); }

    operator float() const
      { return Float(); }

    operator double() const
      { return Double(); }

    operator MD_DateTime() const
      { return DateTime(); }

    operator MD_EpochNS() const
      { return EpochNS(); }
    
    operator std::string() const
      { return String(); }

    std::string ToString() const
      { return CastString(); }

    //FIXME: Finish this
    void FromString(const std::string& TypeStr, const std::string& ValueStr);

    double ToDouble() const
      { return CastDouble(); }

    int ToInt() const
      { return CastInt(); }

    unsigned int ToUnsignedInt() const
      { return CastUnsignedInt(); }

    int64_t ToInt64() const
      { return CastInt64(); }

    uint64_t ToUnsignedInt64() const
      { return CastUnsignedInt64(); }

    MD_DateTime ToDateTime() const
      { return CastDateTime(); }

#endif

    // All constructors take a copy of the data
    static MD_Variant CreateChar(char c);
    static MD_Variant CreateShort(short s);
    static MD_Variant CreateInt(int i);
    static MD_Variant CreateInt64(int64_t i);
    static MD_Variant CreateFloat(float f);
    static MD_Variant CreateDouble(double d);
    static MD_Variant CreateDateTime(MD_DateTime dt);
    static MD_Variant CreateEpochNS(MD_EpochNS et);
    static MD_Variant CreateString(const std::string str);
    static MD_Variant CreateBuffer(const BufferHolder& buf);
    static MD_Variant CreateBuffer(const char *buffer, unsigned int size);
    static MD_Variant CreateMap(const std::map<MD_Variant,MD_Variant>& map);
    static MD_Variant CopyRaw(const char *buffer,unsigned int size, int type);
    static MD_Variant CreateUnsignedInt(unsigned int u);
    static MD_Variant CreateUnsignedInt64(uint64_t u);

    // These functions are used to set the variant.  They will destroy
    // whatever is currently in the variant and replace it with the
    // parameter, setting type appropriately
    void Char(char c);
    void Short(short s);
    void Short(const void * ps);
    void Int(int i);
    void Int(const void * pi);
    void Int64(int64_t i);
    void Int64(const void * pi);
    void UnsignedInt(unsigned int u);
    void UnsignedInt(const void * pu);
    void UnsignedInt64(uint64_t u);
    void UnsignedInt64(const void * pu);
    void Float(float f);
    void Float(const void * pf);
    void Double( double d);
    void Double( const void * pd);
    void DateTime(MD_DateTime dt);
    void DateTime(const void *pdt);
    void EpochNS(const MD_EpochNS &et);
    void EpochNS(const void *pdt);
    void String(const std::string& str);
    void String(const char * str);
    void String(const char * str, size_t iSize);
    void Buffer(const BufferHolder& buff);
    void Buffer(const void * data, size_t iSize);
    void Map(const std::map<MD_Variant, MD_Variant>& map);

    // overloaded comparision operators
    bool operator==(const MD_Variant& rhs) const;
    bool operator!=(const MD_Variant& rhs) const;
    bool operator< (const MD_Variant& rhs) const;
    bool operator> (const MD_Variant& rhs) const;
    bool operator<=(const MD_Variant& rhs) const;
    bool operator>=(const MD_Variant& rhs) const;

    // These are functions that are called by the C and C++ cast
    // operators First they check to make sure that the variant you
    // are casting is of the type that you are trying to cast it too,
    // if not it asserts.
    char Char() const {
      CheckTypeOrThrow(CHAR);
      return data.cdata;
    }

    const int& Int() const {	
      CheckTypeOrThrow(INT);
      return data.idata;
    }

    const int64_t& Int64() const {	
      CheckTypeOrThrow(INT64);
      return data.i64data;
    }

    unsigned int UnsignedInt() const {
      CheckTypeOrThrow(UINT);
      return static_cast<unsigned int>(data.idata);
    }

    uint64_t UnsignedInt64() const {
      CheckTypeOrThrow(UINT64);
      return static_cast<uint64_t>(data.i64data);
    }

    short Short() const {
      CheckTypeOrThrow(SHORT);
      return data.sdata;
    }

    float Float() const {
      CheckTypeOrThrow(FLOAT);
      return data.fdata;
    }

    double Double() const {
      CheckTypeOrThrow(DOUBLE);
      return data.ddata;
    }

    MD_DateTime DateTime() const {
      CheckTypeOrThrow(DATETIME);
      return *(data.dtdata);
    }
    
    const MD_EpochNS& EpochNS() const {
      CheckTypeOrThrow(EPOCH_NSEC);
      return data.etdata;
    }

    const std::string& String() const {
      CheckTypeOrThrow(STRING);
      return *(data.strdata);
    }

    const BufferHolder& Buffer() const {
      CheckTypeOrThrow(BUFFER);
      return *(data.rawdata);
    }

    const std::map<MD_Variant, MD_Variant>& Map() const {
      CheckTypeOrThrow(MAP);
      return *(data.mapdata);
    }

    // These functions will try to cast the Variant to the specified
    // type if possible.  Otherwise they will throw a BadVariantCast

    char CastChar() const;
    int CastInt() const;
    unsigned int CastUnsignedInt() const;
    int64_t CastInt64() const;
    uint64_t CastUnsignedInt64() const;
    short CastShort() const;
    float CastFloat() const;
    double CastDouble() const;
    MD_DateTime CastDateTime() const;
    const std::string CastString() const;
    const BufferHolder CastBuffer() const;
    const std::map<MD_Variant, MD_Variant> CastMap() const;

    int GetType() const
      { return type; }
      	
  private:
    // This is the union that holds the Data Types
    union DataRep {
      char cdata;
      short sdata;
      int idata;
      int64_t i64data;
      float fdata;
      double ddata;
      MD_DateTime *dtdata;
      MD_EpochNS etdata;
      BufferHolder *rawdata;
      std::string *strdata;
      std::map<MD_Variant, MD_Variant> *mapdata;
    };

    inline void CheckTypeOrThrow(VarType t) const {
      if (t != type) {
        //throw DynamicTypeError();
        assert(false);
      }
    }

    void DestroyExisting();

    // Data members of MD_Variant
    VarType type;
    DataRep data;
  };

  inline std::ostream& operator<<(std::ostream& os,
                                  const MD::MD_Variant& variant) {
    switch (variant.GetType()) {
    case CHAR: os << variant.CastChar(); break;
    case INT: os << variant.Int(); break;
    case INT64: os << variant.Int64(); break;
    case UINT: os << variant.UnsignedInt(); break;
    case UINT64: os << variant.UnsignedInt64(); break;
    case SHORT: os << variant.Short(); break;
    case FLOAT: os << variant.Float(); break;
    case DOUBLE: os << frac_double_to_string(variant.Double()); break;
    case STRING: os << variant.String().c_str(); break;
    case DATETIME: os << variant.DateTime(); break;
    case EPOCH_NSEC: os << variant.EpochNS().get(); break; // always return nanosecond in variant
    case MAP: os << "<MAP>"; break;
    case BUFFER: 
    {
        MD::BufferHolder holder = variant.Buffer();
        if( holder.size == 0 || !holder.data )
        {
            os << "<BUFFER>";
        }
        else
        {
            std::string str( (const char*)holder.data, holder.size);
            os << str.c_str();
        }
        break;
    }
    default: os << "<VARIANT>"; break;
    };
    return os;
  }

  inline const MD_Variant operator+(const MD_Variant& lOp,
                                    const MD_Variant& rOp) {
    switch (lOp.GetType()) {
    case INT:
      return MD_Variant(lOp.Int() + rOp.Int()); break;
    case INT64:
      return MD_Variant(lOp.Int64() + rOp.Int64()); break;
    case UINT:
      return MD_Variant(lOp.UnsignedInt() + rOp.UnsignedInt()); break;
    case UINT64:
      return MD_Variant(lOp.UnsignedInt64() + rOp.UnsignedInt64()); break;
    case FLOAT:
      return MD_Variant(lOp.Float() + rOp.Float()); break;
    case DOUBLE:
      return MD_Variant(lOp.Double() + rOp.Double()); break;
    default:
      return MD_Variant(); break;
    }
  }
 
  inline const MD_Variant operator-(const MD_Variant& lOp,
                                    const MD_Variant& rOp) {
    switch (lOp.GetType()) {
    case INT:
      return MD_Variant(lOp.Int() - rOp.Int()); break;
    case INT64:
      return MD_Variant(lOp.Int64() - rOp.Int64()); break;
    case UINT:
      return MD_Variant(lOp.UnsignedInt() - rOp.UnsignedInt()); break;
    case UINT64:
      return MD_Variant(lOp.UnsignedInt64() - rOp.UnsignedInt64()); break;
    case FLOAT:
      return MD_Variant(lOp.Float() - rOp.Float()); break;
    case DOUBLE:
      return MD_Variant(lOp.Double() - rOp.Double()); break;
    default:
      return MD_Variant(); break;
    }
  }
 
  inline const MD_Variant operator*(const MD_Variant& lOp,
                                    const MD_Variant& rOp) {
    switch (lOp.GetType()) {
    case INT:
      return MD_Variant( lOp.Int() * rOp.Int() ); break;
    case INT64:
      return MD_Variant( lOp.Int64() * rOp.Int64() ); break;
    case UINT:
      return MD_Variant( lOp.UnsignedInt() * rOp.UnsignedInt() ); break;
    case UINT64:
      return MD_Variant( lOp.UnsignedInt64() * rOp.UnsignedInt64() ); break;
    case FLOAT:
      return MD_Variant( lOp.Float() * rOp.Float() ); break;
    case DOUBLE:
      return MD_Variant( lOp.Double() * rOp.Double() ); break;
    default:
      return MD_Variant(); break;
    }
  }

  inline const MD_Variant operator/(const MD_Variant& lOp,
                                    const MD_Variant& rOp) {
    switch (lOp.GetType()) {
    case INT:
      return MD_Variant( lOp.Int() / rOp.Int() ); break;
    case INT64:
      return MD_Variant( lOp.Int64() / rOp.Int64() ); break;
    case UINT:
      return MD_Variant( lOp.UnsignedInt() / rOp.UnsignedInt() ); break;
    case UINT64:
      return MD_Variant( lOp.UnsignedInt64() / rOp.UnsignedInt64() ); break;
    case FLOAT:
      return MD_Variant( lOp.Float() / rOp.Float() ); break;
    case DOUBLE:
      return MD_Variant( lOp.Double() / rOp.Double() ); break;
    default:
      return MD_Variant(); break;
    }
  }


  inline const MD_Variant operator%(const MD_Variant& lOp,
                                    const MD_Variant& rOp) {
    switch (lOp.GetType()) {
    case INT:
      return MD_Variant( lOp.Int() / rOp.Int() ); break;
    case INT64:
      return MD_Variant( lOp.Int64() / rOp.Int64() ); break;
    case UINT:
      return MD_Variant( lOp.UnsignedInt() / rOp.UnsignedInt() ); break;
    case UINT64:
      return MD_Variant( lOp.UnsignedInt64() / rOp.UnsignedInt64() ); break;
    default:
      return MD_Variant(); break;
    }
  }

  template <class T>
    inline T typed_static_cast(const MD_Variant& v);

  template <>
    inline char typed_static_cast<char>(const MD_Variant& v)
    { return v.Char(); }

  template <>
    inline int typed_static_cast<int>(const MD_Variant& v)
    { return v.Int(); }

  template <>
    inline unsigned int typed_static_cast<unsigned int>(const MD_Variant& v)
    { return v.UnsignedInt(); }

  template <>
    inline int64_t typed_static_cast<int64_t>(const MD_Variant& v)
    { return v.Int64(); }

  template <>
    inline uint64_t typed_static_cast<uint64_t>(const MD_Variant& v)
    { return v.UnsignedInt64(); }

  template <>
    inline short typed_static_cast<short>(const MD_Variant& v)
    { return v.Short(); }

  template <>
    inline float typed_static_cast<float>(const MD_Variant& v)
    { return v.Float(); }

  template <>
    inline double typed_static_cast<double>(const MD_Variant& v)
    { return v.Double(); }

  template <>
    inline MD_DateTime typed_static_cast<MD_DateTime>(const MD_Variant& v)
    { return v.DateTime(); }

  template <>
    inline std::string typed_static_cast<std::string>(const MD_Variant& v)
    { return v.String(); }

  template <> inline const std::map<MD_Variant, MD_Variant>& typed_static_cast<const std::map <MD_Variant, MD_Variant>& >(const MD_Variant& v)
    { return v.Map(); }

  template <> inline const BufferHolder& typed_static_cast<const BufferHolder& >(const MD_Variant& v)
    { return v.Buffer(); }


  template <class T>
    inline T typed_cast(const MD_Variant& v);

  template <>
    inline char typed_cast<char>(const MD_Variant& v) 
    { return v.CastChar(); }

  template <>
    inline int typed_cast<int>(const MD_Variant& v) 
    { return v.CastInt(); }

  template <>
    inline unsigned int typed_cast<unsigned int>(const MD_Variant& v) 
    { return v.CastUnsignedInt(); }

  template <>
    inline int64_t typed_cast<int64_t>(const MD_Variant& v) 
    { return v.CastInt64(); }

  template <>
    inline uint64_t typed_cast<uint64_t>(const MD_Variant& v) 
    { return v.CastUnsignedInt64(); }

  template <>
    inline short typed_cast<short>(const MD_Variant& v) 
    { return v.CastShort(); }

  template <>
    inline float typed_cast<float>(const MD_Variant& v) 
    { return v.CastFloat(); }

  template <>
    inline double typed_cast<double>(const MD_Variant& v) 
    { return v.CastDouble(); }

  template <>
    inline MD_DateTime typed_cast<MD_DateTime>(const MD_Variant &v)
    { return v.CastDateTime(); }

  template <>
    inline std::string typed_cast<std::string>(const MD_Variant& v) 
    { return v.CastString(); }

  template <>
    inline std::map<MD_Variant, MD_Variant> typed_cast<std::map<MD_Variant,MD_Variant> >(const MD_Variant& v) 
    { return v.CastMap(); }

  template <>
    inline BufferHolder typed_cast<BufferHolder>(const MD_Variant& v) 
    { return v.CastBuffer(); }
}

#endif // __MD_LIB_VARIANT_H__
