#ifndef __MD_LIB_MINT_H__
#define __MD_LIB_MINT_H__

#include <cassert>
#include <string>

#include "MD/MD_Message.h"
#include "MD/MD_DateTime.h"
#include "MD/MD_numeric.h"
#include "MD/MD_Types.h"
#include "MD/MD_Protocols.h"
#include "MD_Logger.h"

namespace MD {

class mInt : public MD_Message
{
  public:
  typedef MD_MsgIterator<mInt> iterator;
  class Field;

  mInt(bool opt = true);
  mInt(bool opt, const MD_Message::Allocator& allocator);
  mInt(const mInt& src);
  mInt(mInt&& src);
  mInt(const unsigned char* ptr, uint32_t size);
  mInt(const unsigned char*         ptr,
       uint32_t                     size,
       const MD_Message::Allocator& allocator);
  virtual ~mInt();

  mInt & operator= (const mInt & src);

  // Give the user a new empty message of the same type as this one.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD_Message> create() const;

  // Give the user a new copy of the message.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD_Message> clone() const;

  // Give the user a new pointer containing this message.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD_Message> move();

  //! Accessor methods
  virtual bool getChar     (FieldID id, char &        result) const;
  virtual bool getInt      (FieldID id, int &         result) const;
  virtual bool getInt64    (FieldID id, int64_t &     result) const;
  virtual bool getFloat    (FieldID id, float &       result) const;
  virtual bool getDouble   (FieldID id, double &      result) const;
  virtual bool getString   (FieldID id, String &      result) const;
  virtual bool getBuffer   (FieldID id, Buffer &      result) const;
  virtual bool getDateTime (FieldID id, MD_DateTime & result) const;
  virtual bool getEpochNS  (FieldID id, MD_EpochNS &  result) const;

  virtual bool addChar     (FieldID id, char        data);
  virtual bool addInt      (FieldID id, int         data);
  virtual bool addInt64    (FieldID id, int64_t     data);
  virtual bool addFloat    (FieldID id, float       data);
  virtual bool addDouble   (FieldID id, double      data);
  virtual bool addString   (FieldID id, String      data);
  virtual bool addBuffer   (FieldID id, Buffer      data);
  virtual bool addDateTime (FieldID id, MD_DateTime data);
  virtual bool addEpochNS  (FieldID id, MD_EpochNS  data);

  virtual bool updateChar     (FieldID id, char        data);
  virtual bool updateInt      (FieldID id, int         data);
  virtual bool updateInt64    (FieldID id, int64_t     data);
  virtual bool updateFloat    (FieldID id, float       data);
  virtual bool updateDouble   (FieldID id, double      data);
  virtual bool updateString   (FieldID id, String      data);
  virtual bool updateBuffer   (FieldID id, Buffer      data);
  virtual bool updateDateTime (FieldID id, MD_DateTime data);
  virtual bool updateEpochNS  (FieldID id, MD_EpochNS  data);

  // Pull in base class convenience methods
  using MD_Message::getString;
  using MD_Message::addString;
  using MD_Message::updateString;

  virtual bool erase (FieldID id);
  virtual bool erase (const std::set<FieldID> &ids);
  virtual bool erase (const std::vector<FieldID> &ids);

  virtual void clear();

  virtual bool empty() const;
  virtual bool exists (FieldID id) const;

  virtual const unsigned char * getBinary() const;
  virtual bool setBinary (const unsigned char * ptr, uint32_t size);

  virtual uint32_t getSize() const;

  virtual PROTOCOLS::MsgProtocol protocol() const;

  virtual const unsigned char *
  getDataPtr (FieldID id, TypeID & type, int & size) const;

  iterator begin() const;
  iterator end()   const;

private:
  virtual void getFieldIDs(std::back_insert_iterator<std::vector<FieldID> > inserter) const;

  template <typename FieldType>
  bool addField (FieldID id, FieldType data);

  template <typename FieldType>
  bool updateField (FieldID id, FieldType data);

  template <typename FieldType>
  bool getHelper(FieldID id, FieldType & data) const;

  static const size_t DATA_START = 1;
};

class mInt::Field
{
  typedef MD::FIELDS::FieldID   FieldID;
  typedef MD::DATATYPES::TypeID TypeID;

  public:
  Field(char * ptr);

  FieldID        id() const;
  TypeID       type() const;
  uint8_t      size() const;
  char    *data_ptr() const;

  operator char()    const;
  operator int32_t() const;
  operator int64_t() const;
  operator float()   const;
  operator double()  const;
  operator String()  const;
  operator Buffer()  const;
  operator MD_EpochNS() const;

  bool operator==(const mInt::Field &rhs) const;

  private:
  static uint8_t header_length();

  char *ptr()      const;
  char *id_ptr()   const;
  char *size_ptr() const;
  char *next_ptr() const;
  bool  isValid() const;
  void  next();
  friend  class mInt;
  friend  class MD_MsgIterator<mInt>;
  friend class MD_Message;
  // state:
  char *m_ptr;
};

#ifndef NO_TEMPLATE_SPECIALIZATION_FORWARD_DEC
// forward declaration of template specializations
template <> bool mInt::getHelper(FieldID id, int32_t & data) const;
template <> bool mInt::addField (FieldID id, int32_t data);
#endif

// Inline function definitions.

inline mInt::mInt(bool opt) : mInt(opt, MD_Message::DEFAULT_ALLOCATOR) {}

inline mInt::mInt(bool opt, const MD_Message::Allocator& allocator)
    : MD_Message(allocator)
{
    m_data->clear();
    m_pOffsets->clear();

    m_data->resize(DATA_START);
    m_data->begin()[0] = PROTOCOLS::MINT;
}

inline mInt::mInt(const mInt& src) : MD_Message()
{
    *this = src;
}

inline mInt::mInt(mInt&& src) : MD_Message()
{
    swap(src);
}

inline mInt::mInt(const unsigned char* ptr, uint32_t size)
    : mInt(ptr, size, MD_Message::DEFAULT_ALLOCATOR)
{}

inline mInt::mInt(const unsigned char*         ptr,
                  uint32_t                     size,
                  const MD_Message::Allocator& allocator)
    : MD_Message(allocator)
{
    setBinary(ptr, size);  // setBinary clears m_data and m_pOffsets
}

inline mInt::~mInt() { }


inline std::unique_ptr<MD_Message> mInt::create() const
{ return std::unique_ptr<MD_Message>(new mInt); }

inline std::unique_ptr<MD_Message> mInt::clone() const {
  return std::unique_ptr<MD_Message>(new mInt(*this));
}

inline std::unique_ptr<MD_Message> mInt::move() {
  return std::unique_ptr<MD_Message>(new mInt(std::move(*this)));
}

inline bool mInt::getChar(FieldID id, char & result) const {
  return getHelper(id, result);
}

inline bool mInt::getInt (FieldID id, int & result) const {
  return getHelper<int32_t>(id, result);
}

inline bool mInt::getInt64 (FieldID id, int64_t & result) const {
  return getHelper(id, result);
}

inline bool mInt::getFloat (FieldID id, float & result) const {
  return getHelper(id, result);
}

inline bool mInt::getDouble (FieldID id, double & result) const {
  return getHelper(id, result);
}

inline bool mInt::getString (FieldID id, String & result) const {
  return getHelper(id, result);
}

inline bool mInt::getBuffer (FieldID id, Buffer & result) const {
  return getHelper(id, result);
}

inline bool mInt::getDateTime (FieldID id, MD_DateTime & result) const {
  return getHelper(id, result);
}

inline bool mInt::getEpochNS (FieldID id, MD_EpochNS & result) const {
  return getHelper(id, result);
}

inline bool mInt::Field::isValid() const { return m_ptr != NULL; }

inline bool mInt::exists (FieldID id) const
{ return getField<Field>(id).isValid(); }

inline bool mInt::addChar(FieldID id, char data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool mInt::addInt (FieldID id, int data) {
  return !exists(id) ? addField<int32_t>(id, data) : false;
}

inline bool mInt::addInt64 (FieldID id, int64_t data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool mInt::addFloat (FieldID id, float data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool mInt::addDouble (FieldID id, double data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool mInt::addString (FieldID id, String data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool mInt::addBuffer (FieldID id, Buffer data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool mInt::addDateTime (FieldID id, MD_DateTime data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool mInt::addEpochNS (FieldID id, MD_EpochNS data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool mInt::updateChar(FieldID id, char data) {
  return updateField(id, data);
}

inline bool mInt::updateInt (FieldID id, int data) {
  return updateField<int32_t>(id, data);
}

inline bool mInt::updateInt64 (FieldID id, int64_t data) {
  return updateField(id, data);
}

inline bool mInt::updateFloat (FieldID id, float data) {
  return updateField(id, data);
}

inline bool mInt::updateDouble (FieldID id, double data) {
  return updateField(id, data);
}

inline bool mInt::updateString (FieldID id, String data) {
  return updateField(id, data);
}

inline bool mInt::updateBuffer (FieldID id, Buffer data) {
  return updateField(id, data);
}

inline bool mInt::updateDateTime (FieldID id, MD_DateTime data) {
  return updateField(id, data);
}

inline bool mInt::updateEpochNS (FieldID id, MD_EpochNS data) {
  return updateField(id, data);
}

inline void mInt::clear() {
  m_data->resize(DATA_START);
  m_pOffsets->clear();
}

inline mInt::iterator mInt::begin() const
{ return mInt::iterator(m_data->begin() + DATA_START); }

inline mInt::iterator mInt::end() const
{ return mInt::iterator(m_data->end()); }

inline uint32_t mInt::getSize() const
{ return (uint32_t)m_data->size(); }

inline bool mInt::empty() const
{ return getSize() <= DATA_START; }

inline const unsigned char * mInt::getBinary() const {
  return reinterpret_cast <unsigned char *> (m_data->begin());
}


inline PROTOCOLS::MsgProtocol mInt::protocol() const {
  return PROTOCOLS::MINT;
}

inline char * mInt::Field::id_ptr()   const { return m_ptr; }

inline MD::FIELDS::FieldID mInt::Field::id() const
{ assert(isValid()); return MD::getint16le(id_ptr()); }

inline char  * mInt::Field::size_ptr() const { return m_ptr + 2; }
inline uint8_t mInt::Field::size() const
{
  assert(isValid());
  return *reinterpret_cast<uint8_t*>(size_ptr());
}
inline char * mInt::Field::data_ptr() const { return m_ptr + 3; }
inline char * mInt::Field::next_ptr() const { return data_ptr() + size(); }

inline void mInt::getFieldIDs(std::back_insert_iterator<std::vector<FieldID> > inserter) const {
  char * ptr = m_data->begin() + DATA_START;
  const char * end_ptr = m_data->end();

  while (ptr < end_ptr) {
    Field field(ptr);
    *inserter++ = field.id();
    ptr = field.next_ptr();
  }
}


// TODO: Make this code resilient to malformed input.  It's
// fragile and prone to buffer overflows, at the moment.
inline mInt::Field::Field (char * ptr) : m_ptr(ptr) {}

inline MD::DATATYPES::TypeID mInt::Field::type() const {
  const MD::MD_FieldMap &inst = MD::MD_FieldMap::instance();
  assert(isValid());
  MD::MD_FieldMap::const_iterator i = inst.findById(id());
  return (i != inst.end()) ? i->type : MD::DATATYPES::RAW;
}

inline char *  mInt::Field::ptr()      const { return m_ptr; }
inline void    mInt::Field::next() { m_ptr = data_ptr() + size(); }
inline uint8_t mInt::Field::header_length() { return 3; }

inline bool mInt::Field::operator==(const mInt::Field &rhs) const
{
  return (type() == rhs.type() && size() == rhs.size() &&
          !memcmp(data_ptr(), rhs.data_ptr(), size()));
}

} // namespace MD

#endif // __MD_LIB_MINT_H__

