#ifndef __MD_LIB_NET_UTILS_H__
#define __MD_LIB_NET_UTILS_H__

#include <string>
#include <assert.h>
#include <cstring>
#include <stddef.h>
#ifdef MD_LIB_LINUX
#include <byteswap.h>
#endif
#include <MD/private/prefix.h>
#include <MD/private/byte_order.h>

#ifdef MD_LIB_WIN
#if defined(_AFX) || defined(_AFXDLL)
#include <afxwin.h>
#endif
#include <winsock2.h>
#else
#include <sys/socket.h>   /* needed for socklen_t        */
#include <netinet/in.h>   /* Needed for ntohl, ntohs ... */
#include <poll.h>
#endif

#include <MD/MD_sys_types.h>
#ifndef MD_LIB_WIN
#include <netinet/in.h>
#endif

// Unaligned memory access routines.  These routines read and write
// potentially misaligned integers of various sizes to and from memory
// buffers using explicitly specified byte orders.  It seems that the
// best ways to implement these routines are VERY platform specific,
// when performance is an issue.  By using these routines you can get
// the fastest implementation for your platform without having to
// litter the rest of your code with platform-specific crud.

namespace MD {

  // Get and return a 64-bit int of a specified byte order from a mem buffer
   int64_t getint64be  (const void *p); // big endian
   int64_t getint64le  (const void *p); // little endian
   int64_t getint64hbo (const void *p); // host byte order
  uint64_t getuint64be (const void *p); // big endian
  uint64_t getuint64le (const void *p); // little endian
  uint64_t getuint64hbo(const void *p); // host byte order

  // Put a 64-bit int into a memory buffer using a specified byte order.
  int setint64be (void *p, int64_t x); // big endian
  int setint64le (void *p, int64_t x); // little endian
  int setint64hbo(void *p, int64_t x); // host byte order

  // Get and return a 32-bit int of a specified byte order from a mem buffer
   int32_t getint32be  (const void *p); // big endian
   int32_t getint32le  (const void *p); // little endian
   int32_t getint32hbo (const void *p); // host byte order
  uint32_t getuint32be (const void *p); // big endian
  uint32_t getuint32le (const void *p); // little endian
  uint32_t getuint32hbo(const void *p); // host byte order

  // Put a 32-bit int into a memory buffer using a specified byte order.
  int setint32be (void *p, int32_t x); // big endian
  int setint32le (void *p, int32_t x); // little endian
  int setint32hbo(void *p, int32_t x); // host byte order

  // Get and return a 16-bit int of a specified byte order from a mem buffer
  int16_t  getint16be  (const void *p); // big endian
  int16_t  getint16le  (const void *p); // little endian
  int16_t  getint16hbo (const void *p); // host byte order
  uint16_t getuint16be (const void *p); // big endian
  uint16_t getuint16le (const void *p); // little endian
  uint16_t getuint16hbo(const void *p); // host byte order

  // Put a 16-bit int into a memory buffer using a specified byte order
  int setint16be (void *p, int16_t x); // big endian
  int setint16le (void *p, int16_t x); // little endian
  int setint16hbo(void *p, int16_t x); // host byte order

  // Get and return a (32-bit) float of a specified byte order from a mem buffer
  float getfloatbe (const void *p);
  float getfloatle (const void *p);
  float getfloathbo(const void *p);

  // Put a (32-bit) float into a memory buffer using a specified byte order
  int setfloatbe (void *p, float x);
  int setfloatle (void *p, float x);
  int setfloathbo(void *p, float x);

  // Get and return a (64-bit) double of a specified byte order from a mem buffer
  double getdoublebe (const void *p);
  double getdoublele (const void *p);
  double getdoublehbo(const void *p);

  // Put a (64-bit) double into a memory buffer using a specified byte order
  int setdoublebe (void *p, double x);
  int setdoublele (void *p, double x);
  int setdoublehbo(void *p, double x);

  // set a std::string, get a std::string
  int         setstringbe(void *p, const std::string &s);
  int         setstringle(void *p, const std::string &s);
  std::string getstringbe(const void *p);
  std::string getstringle(const void *p);

  // parsing routines, return the number of bytes parsed and fill in the pointer
  int parse_stringbe(const void *p, std::string *s);
  int parse_stringle(const void *p, std::string *s);
  int parse_uint16be(const void *p, uint16_t    *i);
  int parse_uint16le(const void *p, uint16_t    *i);
  int parse_int16be (const void *p, int16_t     *i);
  int parse_int16le (const void *p, int16_t     *i);
  int parse_uint32be(const void *p, uint32_t    *i);
  int parse_uint32le(const void *p, uint32_t    *i);
  int parse_int32be (const void *p, int32_t     *i);
  int parse_int32le (const void *p, int32_t     *i);
  int parse_uint64be(const void *p, uint64_t    *i);
  int parse_uint64le(const void *p, uint64_t    *i);
  int parse_int64be (const void *p, int64_t     *i);
  int parse_int64le (const void *p, int64_t     *i);
  int parse_floatbe (const void *p, float       *f);
  int parse_floatle (const void *p, float       *f);
  int parse_doublebe(const void *p, double      *d);
  int parse_doublele(const void *p, double      *d);

  // makes sure that the parsing does not read off the end, returns bytes parsed
  // which can be zero (if there is not enough bytes to read)
  int parse_stringbe(const void *p, const void *e, std::string *s);

  // byteswapping instructions
  int64_t bswap64(int64_t x); // swap regardless of endian
  int32_t bswap32(int32_t x); // swap regardless of endian
  int16_t bswap16(int16_t x); // swap regardless of endian

  // folded int support
  int parse_folded_uint(const void *p, size_t size, uint32_t *i);
  int parse_folded_int (const void *p, size_t size, int32_t  *i);
} // namespace

//
// Implementation
//
namespace MD {

inline int64_t getint64be(const void *p)
{
  return (int64_t)((uint8_t *)p)[7]
    |   ((int64_t)((uint8_t *)p)[6] << 8)
    |   ((int64_t)((uint8_t *)p)[5] << 16)
    |   ((int64_t)((uint8_t *)p)[4] << 24)
    |   ((int64_t)((uint8_t *)p)[3] << 32)
    |   ((int64_t)((uint8_t *)p)[2] << 40)
    |   ((int64_t)((uint8_t *)p)[1] << 48)
    |   ((int64_t)((uint8_t *)p)[0] << 56);
}

inline uint64_t getuint64be(const void *p)
{ return (uint64_t)getint64be(p); }

inline int64_t getint64le(const void *p)
{
#if defined(MD_LIB_UNALIGNED_ACCESS) && (MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN)
  return *(int64_t *)p;
#else
  return (int64_t)((uint8_t *)p)[0]
    |   ((int64_t)((uint8_t *)p)[1] << 8)
    |   ((int64_t)((uint8_t *)p)[2] << 16)
    |   ((int64_t)((uint8_t *)p)[3] << 24)
    |   ((int64_t)((uint8_t *)p)[4] << 32)
    |   ((int64_t)((uint8_t *)p)[5] << 40)
    |   ((int64_t)((uint8_t *)p)[6] << 48)
    |   ((int64_t)((uint8_t *)p)[7] << 56);
#endif
}

inline uint64_t getuint64le(const void *p)
{ return (uint64_t)getint64le(p); }

inline int64_t getint64hbo(const void *p)
{
#if MD_LIB_BYTE_ORDER == MD_LIB_BIG_ENDIAN
  return getint64be(p);
#elif MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN
  return getint64le(p);
#else
#error "Invalid MD_LIB_BYTE_ORDER"
#endif
}

inline uint64_t getuint64hbo(const void *p)
{ return (uint64_t)getint64hbo(p); }

inline int setint64be(void *p, int64_t x)
{
  ((uint8_t *)p)[7] = (uint8_t)x;
  ((uint8_t *)p)[6] = (uint8_t)(x >> 8);
  ((uint8_t *)p)[5] = (uint8_t)(x >> 16);
  ((uint8_t *)p)[4] = (uint8_t)(x >> 24);
  ((uint8_t *)p)[3] = (uint8_t)(x >> 32);
  ((uint8_t *)p)[2] = (uint8_t)(x >> 40);
  ((uint8_t *)p)[1] = (uint8_t)(x >> 48);
  ((uint8_t *)p)[0] = (uint8_t)(x >> 56);
  return 8;
}

inline int setint64le(void *p, int64_t x)
{
#if (MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN)
  memcpy(p, &x, sizeof(x));
#else
  ((uint8_t *)p)[0] = (uint8_t)x;
  ((uint8_t *)p)[1] = (uint8_t)(x >> 8);
  ((uint8_t *)p)[2] = (uint8_t)(x >> 16);
  ((uint8_t *)p)[3] = (uint8_t)(x >> 24);
  ((uint8_t *)p)[4] = (uint8_t)(x >> 32);
  ((uint8_t *)p)[5] = (uint8_t)(x >> 40);
  ((uint8_t *)p)[6] = (uint8_t)(x >> 48);
  ((uint8_t *)p)[7] = (uint8_t)(x >> 56);
#endif
  return 8;
}

inline int setint64hbo(void *p, int64_t x)
{
#if MD_LIB_BYTE_ORDER == MD_LIB_BIG_ENDIAN
  return setint64be(p, x);
#elif MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN
  return setint64le(p, x);
#else
#error "Invalid MD_LIB_BYTE_ORDER"
#endif
}

inline int32_t getint32be(const void *p)
{
#if defined(MD_LIB_UNALIGNED_ACCESS)
  return ntohl((*(int32_t *)p));
#else
  return ((uint8_t *)p)[3]
    |   (((uint8_t *)p)[2] << 8)
    |   (((uint8_t *)p)[1] << 16)
    |   (((uint8_t *)p)[0] << 24);
#endif
}

inline uint32_t getuint32be(const void *p)
{ return (uint32_t)getint32be(p); }

inline int32_t getint32le(const void *p)
{
#if defined(MD_LIB_UNALIGNED_ACCESS) && (MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN)
  return *(int32_t *)p;
#else
  return ((uint8_t *)p)[0]
    |   (((uint8_t *)p)[1] << 8)
    |   (((uint8_t *)p)[2] << 16)
    |   (((uint8_t *)p)[3] << 24);
#endif
}

inline uint32_t getuint32le(const void *p)
{ return (uint32_t)getint32le(p); }

inline int32_t getint32hbo(const void * p)
{
#if MD_LIB_BYTE_ORDER == MD_LIB_BIG_ENDIAN
  return getint32be(p);
#elif MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN
  return getint32le(p);
#else
#error "Invalid MD_LIB_BYTE_ORDER"
#endif
}

inline uint32_t getuint32hbo(const void * p)
{ return (uint32_t)getint32hbo(p); }

inline int parse_int32be(const void *p, int32_t *i)
{
  *i = getint32be(p);
  return 4;
}
inline int parse_uint32be(const void *p, uint32_t *i)
{ return parse_int32be(p, (int32_t *)i); }

inline int parse_int64be(const void *p, int64_t *i)
{
  *i = getint64be(p);
  return 8;
}
inline int parse_uint64be(const void *p, uint64_t *i)
{ return parse_int64be(p, (int64_t *)i); }

inline int parse_int32le(const void *p, int32_t *i)
{
  *i = getint32le(p);
  return 4;
}
inline int parse_uint32le(const void *p, uint32_t *i)
{ return parse_int32le(p, (int32_t *)i); }

inline int parse_int64le(const void *p, int64_t *i)
{
  *i = getint64le(p);
  return 8;
}
inline int parse_uint64le(const void *p, uint64_t *i)
{ return parse_int64le(p, (int64_t *)i); }

inline int setint32be(void *p, int32_t x)
{
#if defined(MD_LIB_UNALIGNED_ACCESS)
  *(int32_t *)p = htonl(x);
#else
  ((uint8_t *)p)[3] = (uint8_t)x;
  ((uint8_t *)p)[2] = (uint8_t)(x >> 8);
  ((uint8_t *)p)[1] = (uint8_t)(x >> 16);
  ((uint8_t *)p)[0] = (uint8_t)(x >> 24);
#endif
  return 4;
}

inline int setint32le(void * p, int32_t x)
{
#if (MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN)
  memcpy(p, &x, sizeof(x));
#else
  ((uint8_t *)p)[0] = (uint8_t)x;
  ((uint8_t *)p)[1] = (uint8_t)(x >> 8);
  ((uint8_t *)p)[2] = (uint8_t)(x >> 16);
  ((uint8_t *)p)[3] = (uint8_t)(x >> 24);
#endif
  return 4;
}

inline int setint32hbo(void *p, int32_t x)
{
#if MD_LIB_BYTE_ORDER == MD_LIB_BIG_ENDIAN
  return setint32be(p, x);
#elif MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN
  return setint32le(p, x);
#else
#error "Invalid MD_LIB_BYTE_ORDER"
#endif
}

inline int16_t getint16be(const void *p)
{
#if defined(MD_LIB_UNALIGNED_ACCESS)
  return ntohs((*(int16_t *)p));
#else
  return ((uint8_t *)p)[1]
    |   (((uint8_t *)p)[0] << 8);
#endif
}

inline uint16_t getuint16be(const void *p)
{ return (uint16_t)getint16be(p); }

inline int16_t getint16le(const void *p)
{
#if defined(MD_LIB_UNALIGNED_ACCESS) && (MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN)
  return *(int16_t *)p;
#else
  return ((uint8_t *)p)[0]
    |   (((uint8_t *)p)[1] << 8);
#endif
}

inline uint16_t getuint16le(const void *p)
{ return (uint16_t)getint16le(p); }

inline int16_t getint16hbo(const void *p)
{
#if MD_LIB_BYTE_ORDER == MD_LIB_BIG_ENDIAN
  return getint16be(p);
#elif MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN
  return getint16le(p);
#else
#error "Invalid MD_LIB_BYTE_ORDER"
#endif
}

inline uint16_t getuint16hbo(const void *p)
{ return (uint16_t)getint16hbo(p); }

inline std::string getstringbe(const void *p)
{
  return std::string(((const char *)p)+2, getuint16be(p));
}

inline std::string getstringle(const void *p)
{
  return std::string(((const char *)p)+2, getuint16le(p));
}

inline int parse_int16be(const void *p, int16_t *i)
{
  *i = getint16be(p);
  return 2;
}

inline int parse_uint16be(const void *p, uint16_t *i)
{ return parse_int16be(p, (int16_t *)i); }

inline int parse_int16le(const void *p, int16_t *i)
{
  *i = getint16le(p);
  return 2;
}

inline int parse_uint16le(const void *p, uint16_t *i)
{ return parse_int16le(p, (int16_t *)i); }

inline int setint16be(void *p, int16_t x)
{
#if defined(MD_LIB_UNALIGNED_ACCESS)
  *(int16_t *)p = htons(x);
#else
  ((uint8_t *)p)[1] = (uint8_t)x;
  ((uint8_t *)p)[0] = (uint8_t)(x >> 8);
#endif
  return 2;
}

inline int setint16le(void *p, int16_t x)
{
#if defined(MD_LIB_UNALIGNED_ACCESS) && (MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN)
  *(int16_t *)p = x;
#else
  ((uint8_t *)p)[0] = (uint8_t)x;
  ((uint8_t *)p)[1] = (uint8_t)(x >> 8);
#endif
  return 2;
}

inline int setint16hbo(void *p, int16_t x)
{
#if MD_LIB_BYTE_ORDER == MD_LIB_BIG_ENDIAN
  return setint16be(p, x);
#elif MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN
  return setint16le(p, x);
#else
#error "Invalid MD_LIB_BYTE_ORDER"
#endif
}

typedef union
{
  int32_t i32;
  float   f32;
} float_union;

inline float getfloatbe(const void *p)
{
  float_union in;
  in.i32 = getint32be(p);
  return in.f32;
}

inline float getfloatle(const void *p)
{
  float_union in;
  in.i32 = getint32le(p);
  return in.f32;
}

inline float getfloathbo(const void *p)
{
  float_union in;
  in.i32 = getint32hbo(p);
  return in.f32;
}

inline int parse_floatbe (const void *p, float *f)
{
  *f = getfloatbe(p);
  return 4;
}
 
inline int parse_floatle (const void *p, float *f)
{
  *f = getfloatle(p);
  return 4;
}

inline int setfloatbe(void *p, float x)
{
  float_union out;
  out.f32 = x;
  return setint32be(p, out.i32);
}

inline int setfloatle(void *p, float x)
{
  float_union out;
  out.f32 = x;
  return setint32le(p, out.i32);
}

inline int setfloathbo(void *p, float x)
{
  float_union out;
  out.f32 = x;
  return setint32hbo(p, out.i32);
}

typedef union
{
  int64_t i64;
  double  d64;
} double_union;

inline double getdoublebe(const void *p)
{
  double_union in;
  in.i64 = getint64be(p);
  return in.d64;
}

inline double getdoublele(const void *p)
{
  double_union in;
  in.i64 = getint64le(p);
  return in.d64;
}

inline double getdoublehbo(const void *p)
{
  double_union in;
  in.i64 = getint64hbo(p);
  return in.d64;
}

inline int parse_doublebe(const void *p, double *d)
{
  *d = getdoublebe(p);
  return 8;
}

inline int parse_doublele(const void *p, double *d)
{
  *d = getdoublele(p);
  return 8;
}

inline int setdoublebe(void *p, double x)
{
  double_union out;
  out.d64 = x;
  return setint64be(p, out.i64);
}

inline int setdoublele(void *p, double x)
{
  double_union out;
  out.d64 = x;
  return setint64le(p, out.i64);
}

inline int setdoublehbo(void *p, double x)
{
  double_union out;
  out.d64 = x;
  return setint64hbo(p, out.i64);
}

inline int setstringbe(void *p, const std::string &s)
{
  setint16be(p, (int16_t)(s.size() & 0xffff));
  memcpy(((char *)p)+2, s.data(), s.size() & 0xffff);
  return 2+((unsigned int)s.size() & 0xffff);
}

inline int setstringle(void *p, const std::string &s)
{
  setint16le(p, (int16_t)(s.size() & 0xffff));
  memcpy(((char *)p)+2, s.data(), s.size() & 0xffff);
  return 2+((unsigned int)s.size() & 0xffff);
}

inline int parse_stringbe(const void *p, const void *e, std::string *s)
{
  ptrdiff_t d=(const char *)e - (const char *)p;
  if (d >= 2) {
    uint16_t str_len = getuint16be(p);
    if (d >= (2+str_len)) {
      s->assign((const char *)p+2, str_len);
      assert(s->size() == str_len);
      return str_len + 2;
    }
  }
  return 0;
}

inline int parse_stringbe(const void *p, std::string *s)
{
  s->assign(((const char *)p)+2, getuint16be(p));
  assert(s->size() == getuint16be(p));
  return (unsigned int)s->size() + 2;
}

inline int parse_stringle(const void *p, std::string *s)
{
  s->assign(((const char *)p)+2, getuint16le(p));
  assert(s->size() == (unsigned int)getuint16le(p));
  return (unsigned int)s->size() + 2;
}

inline int64_t bswap64(int64_t x)
{
  return ( (x << 56) & 0xff00000000000000ULL ) |
         ( (x << 40) & 0x00ff000000000000ULL ) |
         ( (x << 24) & 0x0000ff0000000000ULL ) |
         ( (x <<  8) & 0x000000ff00000000ULL ) |
         ( (x >>  8) & 0x00000000ff000000ULL ) |
         ( (x >> 24) & 0x0000000000ff0000ULL ) |
         ( (x >> 40) & 0x000000000000ff00ULL ) |
         ( (x >> 56) & 0x00000000000000ffULL );
}

inline int32_t bswap32(int32_t x)
{
  return ( (x << 24) & 0xff000000 ) |
         ( (x <<  8) & 0x00ff0000 ) |
         ( (x >>  8) & 0x0000ff00 ) |
         ( (x >> 24) & 0x000000ff );
}

inline int16_t bswap16(int16_t x)
{
  return (int16_t)(( (x <<  8) & 0xff00 ) |
                   ( (x >>  8) & 0x00ff ));
}

/*
inline int parse_folded_uint(void *p, size_t size, uint32_t x)
{
  unsigned char *pp = (unsigned char *)p;
  if (size && (!(pp[0] & 0x80))) {
    *x = *pp;
    return 1;
  }
  if (size > 1 && (!(pp[0] & 0x40)) {
    *x = ((pp[0] << 8 | pp[1]) & 0x3fff);
    return 2;
  }
  else if (size > 2 && (!(pp[0] & 0x20)) {
    *x = ((pp[0] << 16 | pp[1] << 8 | pp[2]) & 0x001fffff);
    return 3;
  }

  // *value = ((pp[0]<<24 | pp[1]<<16 | pp[2]<<8 | pp[3]) & 0x1fffffff);
  *x = (pp[0]<<24 | pp[1]<<16 | pp[2]<<8 | pp[3]);
  if (*value != 0xffffffff)
    *value &= 0x1fffffff;
  return 4;
}
*/

} // namespace

#ifndef MD_LIB_LINUX
#define __bswap_64(x) MD::bswap64((x))
#endif

# if MD_LIB_BYTE_ORDER == MD_LIB_BIG_ENDIAN
/* The host byte order is the same as network byte order,
   so these functions are all just identity.  */
# define md_ntohll(x)   (x)
# define md_htonll(x)   (x)
# else
#  if MD_LIB_BYTE_ORDER == MD_LIB_LITTLE_ENDIAN
#   define md_ntohll(x) __bswap_64 (x)
#   define md_htonll(x) __bswap_64 (x)
#  endif
# endif
#endif

