#include <boost/algorithm/string.hpp>
#include <boost/program_options.hpp>
#include <iomanip>
#include <iostream>
#include "BroadcastCLI.h"
#include "BroadcastFields.h"
#include "BroadcastMessageTypes.h"
#include "Error.h"

namespace {
const std::string PROMPT = "> ";
const std::string HELP = "HELP";
const std::string CREATE = "CREATE";
const std::string DESTROY = "DESTROY";
const std::string AUTHENTICATE = "AUTHENTICATE";
const std::string CONNECT = "CONNECT";
const std::string DISCONNECT = "DISCONNECT";
const std::string LOGIN = "LOGIN";
const std::string SUBSCRIBE = "SUBSCRIBE";
const std::string IS_SUBSCRIBED = "IS_SUBSCRIBED";
const std::string UNSUBSCRIBE = "UNSUBSCRIBE";
const std::string TIME_RECOVER = "TIME_RECOVER";
const std::string SEQ_RECOVER = "SEQ_RECOVER";
const std::string VERBOSE_LOGGING = "VERBOSE_LOGGING";
const std::string QUIT = "QUIT";
const std::string STATS = "STATS";
const std::string STATS_ONCE = "STATS_ONCE";
const std::string SLEEP = "SLEEP";
const std::string REQUEST_FILE = "REQUEST_FILE";
const std::string SET_CUSTOM_LOGGING = "SET_CUSTOM_LOGGING";
const std::string SET_LATENCY_TRACKING = "SET_LATENCY_TRACKING";

std::atomic<uint64_t> G_int_count(0);
std::atomic<uint64_t> G_int64_count(0);
std::atomic<uint64_t> G_double_count(0);
std::atomic<uint64_t> G_string_count(0);
std::atomic<uint64_t> G_char_count(0);
std::atomic<uint64_t> G_ns_count(0);
}  // namespace

thread_local std::map<std::string, int> BroadcastCLI::m_sequence_map;

static void on_msg(const std::string& topic,
                   FEConsumer::MsgType& msg,
                   void* closure)
{
    BroadcastCLI* app = static_cast<BroadcastCLI*>(closure);
    if(app) {
        app->on_message(topic, msg);
    }
}

BroadcastCLI::BroadcastCLI()
    : m_continue_running(true)
    , m_continue_dispatching(false)
    , m_console_logging(false)
    , m_print_gaps(false)
    , m_print_messages(false)
    , m_current_latency(0)
    , m_highest_latency(0)
    , m_msg_count(0)
    , m_last_msg_count(0)
    , m_gaps_count(0)
    , m_last_gaps_count(0)
    , m_bytes_count(0)
    , m_last_bytes_count(0)
    , m_updates(1)
    , m_stats_timer(m_io_service)
    , m_print_stats(false)
    , m_field_map(MD::MD_FieldMap::instance())
    , m_stats_thread(std::bind(&BroadcastCLI::start_stats_thread, this))
{}

BroadcastCLI::~BroadcastCLI()
{
    quit();
    if(m_stats_thread.joinable()) {
        m_stats_thread.join();
    }
}

void BroadcastCLI::start(bool console_logging,
                         bool print_gaps,
                         bool print_messages,
                         bool iterate_messages,
                         std::vector<std::string>& inputs,
                         const std::vector<std::string>& dispatch_prefixes,
                         unsigned long slow_threshold)
{
    m_console_logging = console_logging;
    if(!m_console_logging) {
        FEConsumer::log_open("./broadcast_cli.log", false);
    }
    m_print_gaps = print_gaps;
    m_print_messages = print_messages;
    m_iterate_messages = iterate_messages;
    m_dispatch_prefixes = dispatch_prefixes;
    m_slow_threshold = slow_threshold;
    if(m_dispatch_prefixes.empty()) {
        m_dispatch_prefixes.push_back("");
    }
    std::string command;
    std::vector<std::string> tokens;
    while(m_continue_running) {
        if(!inputs.empty()) {
            command = inputs.front();
            std::cout << PROMPT << command << std::endl;
            inputs.erase(inputs.begin());
        } else {
            std::cout << PROMPT;
            std::getline(std::cin, command);
        }
        tokens.clear();
        boost::split(tokens, command, boost::is_any_of(" "));
        if(boost::iequals(tokens[0], HELP)) {
            print_help();
        } else if(boost::iequals(tokens[0], QUIT)) {
            quit();
        } else if(boost::iequals(tokens[0], CREATE)) {
            if(tokens.size() < 3) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            int subscribers, workers;
            subscribers = std::atoi(tokens[1].c_str());
            workers = std::atoi(tokens[2].c_str());
            if(subscribers < 1) {
                subscribers = 1;
            }
            if(workers < 1) {
                workers = 1;
            }
            /*bool compress = true;
            if(tokens.size() > 3 && boost::iequals(tokens[3],
            "compression_off")) { compress = false;
            }*/
            create(subscribers, workers);
        } else if(boost::iequals(tokens[0], DESTROY)) {
            destroy();
        } else if(boost::iequals(tokens[0], AUTHENTICATE)) {
            if(tokens.size() < 6) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            authenticate(tokens[1], tokens[2], tokens[3], tokens[4], tokens[5]);
        } else if(boost::iequals(tokens[0], CONNECT)) {
            connect();
        } else if(boost::iequals(tokens[0], DISCONNECT)) {
            disconnect();
        } else if(boost::iequals(tokens[0], LOGIN)) {
            bool update = true;
            if(tokens.size() > 1 && boost::iequals(tokens[1], "false")) {
                update = false;
            }
            login(update);
        } else if(boost::iequals(tokens[0], SUBSCRIBE)) {
            if(tokens.size() < 2) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            bool snap_required = false;
            FEConsumer::DataMode dm = FEConsumer::DELAYED;
            if(tokens.size() > 2 && boost::iequals(tokens[2], "true")) {
                snap_required = true;
            }
            if(tokens.size() > 3) {
                if(boost::iequals(tokens[3], "canned")) {
                    dm = FEConsumer::CANNED;
                } else if(boost::iequals(tokens[3], "realtime")) {
                    dm = FEConsumer::LIVE;
                }
            }
            subscribe(tokens[1], snap_required, dm);
        } else if(boost::iequals(tokens[0], IS_SUBSCRIBED)) {
            if(tokens.size() < 2) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            is_subscribed(tokens[1]);
        } else if(boost::iequals(tokens[0], UNSUBSCRIBE)) {
            if(tokens.size() < 2) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            unsubscribe(tokens[1]);
        } else if(boost::iequals(tokens[0], REQUEST_FILE)) {
            if(tokens.size() < 2) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            request_file(tokens[1]);
        } else if(boost::iequals(tokens[0], TIME_RECOVER)) {
            if(tokens.size() < 4) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            // TODO
        } else if(boost::iequals(tokens[0], SEQ_RECOVER)) {
            if(tokens.size() < 4) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            recover(tokens[1],
                    std::atoi(tokens[2].c_str()),
                    std::atoi(tokens[3].c_str()));
        } else if(boost::iequals(tokens[0], VERBOSE_LOGGING)) {
            if(tokens.size() < 2) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            bool enable = boost::iequals(tokens[1], "true");
            if(enable) {
                FEConsumer::enable_verbose_logging();
                std::cout << "Verbose logging enabled" << std::endl;
            } else {
                FEConsumer::disable_verbose_logging();
                std::cout << "Verbose logging disable" << std::endl;
            }
        } else if(boost::iequals(tokens[0], STATS)) {
            std::unique_lock<std::mutex> guard(m_lock);
            m_print_stats = true;
            while(m_print_stats) {
                m_stats_cv.wait(guard);
            }
            std::cout << std::endl;
        } else if(boost::iequals(tokens[0], STATS_ONCE)) {
            print_stats();
        } else if(boost::iequals(tokens[0], SLEEP)) {
            if(tokens.size() < 2) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            int seconds = std::atoi(tokens[1].c_str());
            std::cout << "Sleeping for " << seconds << " seconds" << std::endl;
            std::this_thread::sleep_for(std::chrono::seconds(seconds));
        } else if(boost::iequals(tokens[0], SET_CUSTOM_LOGGING)) {
            if(tokens.size() < 2) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            set_custom_logging(boost::iequals(tokens[1], "true"));
        } else if(boost::iequals(tokens[0], SET_LATENCY_TRACKING)) {
            if(tokens.size() < 2) {
                print_too_few_args();
                print_suggest_help();
                continue;
            }
            unsigned long bucket_s = 1;
            if(tokens.size() > 2) {
                bucket_s = std::atoi(tokens[2].c_str());
                if(!bucket_s) {
                    bucket_s = 1;
                }
            }
            set_latency_tracking(boost::iequals(tokens[1], "true"), bucket_s);
        } else {
            std::cout << "Unknown command '" << tokens[0] << "'" << std::endl;
            print_suggest_help();
        }
    }
}

bool BroadcastCLI::sigint_handler()
{
    std::lock_guard<std::mutex> guard(m_lock);
    if(m_print_stats) {
        m_print_stats = false;
        m_stats_cv.notify_one();
        return false;
    }
    return true;
}

void BroadcastCLI::start_stats_thread()
{
    boost::asio::io_service::work work(m_io_service);
    m_stats_timer.expires_from_now(boost::posix_time::seconds(1));
    m_stats_timer.async_wait(
        std::bind(&BroadcastCLI::update_stats, this, std::placeholders::_1));
    m_io_service.run();
}

void BroadcastCLI::print_help()
{
    std::cout
        << "Valid commands:" << std::endl
        << std::endl
        << "help" << std::endl
        << "\tPrints this message" << std::endl
        << std::endl
        << "quit" << std::endl
        << "\tExits this application" << std::endl
        << std::endl
        << "create <subscriber thread count> <worker thread count>" << std::endl
        << "\tCreates a new consumer with the given number of subscriber "
           "and worker threads"
        << std::endl
        << std::endl
        << "destroy" << std::endl
        << "\tDestroys a previously created consumer" << std::endl
        << std::endl
        << "authenticate <auth server host> <user> <serial> <device id> <path>"
        << std::endl
        << "\tAuthenticates with the auth server at the given address "
           "using the given credentials"
        << std::endl
        << "\tPath indicates the location of a file named <device "
           "id>.data containing the key and counter"
        << std::endl
        << std::endl
        << "connect" << std::endl
        << "\tConnects to the broadcast server after authenticating"
        << std::endl
        << std::endl
        << "disconnect" << std::endl
        << "\tCompletely disconnects from the broadcast server" << std::endl
        << std::endl
        << "login [update field_map.txt (true|false)]" << std::endl
        << "\tLogs in to the broadcast server after connecting" << std::endl
        << "\tThe value of update field_map.txt defaults to true if omitted"
        << std::endl
        << std::endl
        << "subscribe <topic> [snapshot required (true|false)] "
           "[data mode (delayed|realtime|canned)]"
        << std::endl
        << "\tSubscribes to the given topic" << std::endl
        << "\tIf snapshot required is omitted, no snapshot will be "
           "requested"
        << std::endl
        << "\tIf data mode is omitted, delayed data will be requested"
        << std::endl
        << std::endl
        << "unsubscribe <topic>" << std::endl
        << "\tUnsubscribes from the given topic" << std::endl
        << std::endl
        << "is_subscribed <topic>" << std::endl
        << "\tIndicates whether the given topic is subscribed" << std::endl
        << std::endl
        << "verbose_logging <true|false>" << std::endl
        << "\tEnables or disables verbose logging" << std::endl
        << std::endl
        << "stats" << std::endl
        << "\tBegin printing message stats until cancelled with ctrl+c"
        << std::endl
        << std::endl
        << "stats_once" << std::endl
        << "\tPrint the latest message stats once" << std::endl
        << std::endl
        << "sleep <seconds>" << std::endl
        << "\tSleeps for the given number of seconds" << std::endl
        << std::endl
        << "set_custom_logging <true|false>" << std::endl
        << "\tEnable or disable custom logging" << std::endl
        << std::endl
        << "set_latency_tracking <true|false> (<bucket interval seconds>)"
        << std::endl
        << "\tEnable or disable latency tracking with the given interval "
           "seconds. If unspecified, the interval will be set to 1 second."
        << std::endl
        << std::endl;
}

void BroadcastCLI::quit()
{
    m_continue_running = false;
    m_io_service.stop();
    if(m_consumer != NULL) {
        destroy();
    }
}

void BroadcastCLI::create(int sub_threads, int worker_threads)
{
    if(m_consumer != NULL) {
        std::cout << "Consumer already exists, use destroy before trying to "
                     "create another"
                  << std::endl;
        return;
    }
    m_consumer.reset(
        new FEConsumer(sub_threads, worker_threads, FEConsumer::HWM_UNLIMITED));
    if(m_consumer == NULL) {
        std::cout << "Failed to create consumer" << std::endl;
    } else {
        std::cout << "Consumer created with " << sub_threads
                  << " subscriber threads and " << worker_threads
                  << " worker threads" /*, and compression "
                   << (compress ? "enabled" : "disabled") */
                  << std::endl;
    }
    m_consumer->timeout_sec(10);
    m_consumer->event_cb(&BroadcastCLI::on_event);
    m_consumer->max_snapshot_queue_size(100000);
    m_continue_dispatching = true;
    for(auto& prefix : m_dispatch_prefixes) {
        int fd;
        Error err = m_consumer->register_callback(prefix, &on_msg, this, fd);
        if(err != Error::NoError) {
            std::cout << "Failed to register callback " << prefix << ": "
                      << err.description() << std::endl;
            destroy();
            return;
        }
        std::cout << "Creating dispatch thread for topics starting with \""
                  << prefix << "\"" << std::endl;
        m_dispatch_threads.emplace_back([=] {
            std::vector<struct pollfd> fds;
            fds.push_back({fd, POLLIN, 0});
            while(m_continue_dispatching) {
                int rc = poll(fds.data(), fds.size(), 1000);
                if(rc < 0) {
                    std::cerr << "Poll failed: " << rc << std::endl;
                }
                if(rc < 1) {
                    continue;
                }
                for(size_t i = 0; i < fds.size(); i++) {
                    if(fds[i].revents & POLLIN) {
                        fds[i].revents = 0;
                        m_consumer->dispatch_fd(fds[i].fd);
                    }
                }
            }
        });
    }
}

void BroadcastCLI::destroy()
{
    if(m_consumer == nullptr) {
        print_consumer_null();
        return;
    }
    m_continue_dispatching = false;
    for(auto& t : m_dispatch_threads) {
        if(t.joinable()) {
            t.join();
        }
    }
    m_dispatch_threads.clear();
    m_consumer = nullptr;
    std::cout << "Consumer destroyed successfully" << std::endl;
}

void BroadcastCLI::authenticate(const std::string& host,
                                const std::string& user,
                                const std::string& serial,
                                const std::string& device_id,
                                const std::string& path)
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    Error err = m_consumer->authenticate(
        host, user, serial, device_id, "", "", path, false);
    if(err != Error::NoError) {
        std::cout << "Error " << err.code()
                  << " while trying to authenticate: " << err.description()
                  << std::endl;
    } else {
        std::cout << "Authenticated successfully" << std::endl;
    }
}

void BroadcastCLI::connect()
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    Error err = m_consumer->connect();
    if(err != Error::NoError) {
        std::cout << "Error " << err.code()
                  << " while trying to connect: " << err.description()
                  << std::endl;
    } else {
        std::cout << "Connected successfully" << std::endl;
    }
}

void BroadcastCLI::disconnect()
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    Error err = m_consumer->disconnect();
    if(err != Error::NoError) {
        std::cout << "Error " << err.code()
                  << " while trying to disconnect: " << err.description()
                  << std::endl;
    } else {
        std::cout << "Disconnected successfully" << std::endl;
    }
}

void BroadcastCLI::login(bool update)
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    Error err = m_consumer->log_in(update);
    if(err != Error::NoError) {
        std::cout << "Error " << err.code()
                  << " while trying to log in: " << err.description()
                  << std::endl;
    } else {
        std::cout << "Logged in successfully" << std::endl;
    }
}

void BroadcastCLI::subscribe(const std::string& topic,
                             bool snapshots_required,
                             FEConsumer::DataMode data_mode)
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    FEConsumer::SnapshotMode snapshot_mode = snapshots_required
                                                 ? FEConsumer::REQUIRE_SNAPSHOT
                                                 : FEConsumer::NO_SNAPSHOT;
    Error err = m_consumer->subscribe(topic, snapshot_mode, data_mode);
    if(err != Error::NoError) {
        std::cout << "Error " << err.code()
                  << " while trying to subscribe: " << err.description()
                  << std::endl;
    } else {
        std::cout << "Subscribed to " << topic << std::endl;
    }
}

void BroadcastCLI::is_subscribed(const std::string& topic)
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    std::string status = m_consumer->is_subscribed(topic) ? "true" : "false";
    std::cout << "Subscribed to " << topic << ": " << status << std::endl;
}

void BroadcastCLI::unsubscribe(const std::string& topic)
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    Error err = m_consumer->unsubscribe(topic);
    if(err != Error::NoError) {
        std::cout << "Error " << err.code()
                  << " while trying to unsubscribe: " << err.description()
                  << std::endl;
    } else {
        std::cout << "Unsubscribed from " << topic << std::endl;
    }
}

void BroadcastCLI::request_file(const std::string& filename)
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    std::string contents;
    Error err = m_consumer->request_file(filename, contents);
    if(err != Error::NoError) {
        std::cout << "Error " << err.code()
                  << " while requesting file: " << err.description()
                  << std::endl;
        return;
    }
    std::cout << filename << ":" << std::endl;
    std::cout << contents << std::endl;
}

void BroadcastCLI::recover(const std::string& topic,
                           time_t begin_time,
                           time_t end_time)
{
    /*if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    Error err = m_consumer->recover(topic, begin_time, end_time);
    if(err != Error::NoError) {
        std::cout << "Error " << err.code() << " while trying to recover: " <<
    err.description() << std::endl;
    }*/
    std::cout << "Not implemented" << std::endl;
}

void BroadcastCLI::recover(const std::string& topic, int begin_seq, int end_seq)
{
    /*if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    Error err = m_consumer->recover(topic, begin_seq, end_seq);
    if(err != Error::NoError) {
        std::cout << "Error " << err.code() << " while trying to recover: " <<
    err.description() << std::endl;
    }*/
    std::cout << "Not implemented" << std::endl;
}

void BroadcastCLI::set_custom_logging(bool enable)
{
    if(enable) {
        FEConsumer::set_logging_callback(&BroadcastCLI::on_logging_message);
    } else {
        FEConsumer::set_logging_callback(NULL);
    }
    std::string word = enable ? "enabled" : "disabled";
    std::cout << "Custom logging " << word << std::endl;
}

void BroadcastCLI::set_latency_tracking(bool enable, unsigned long bucket_s)
{
    if(m_consumer == NULL) {
        print_consumer_null();
        return;
    }
    if(enable) {
        m_consumer->enable_latency_tracking(bucket_s);
        std::cout << "Enable latency tracking with " << bucket_s
                  << "s bucket interval" << std::endl;
    } else {
        m_consumer->disable_latency_tracking();
        std::cout << "Disabled latency tracking" << std::endl;
    }
}

void BroadcastCLI::on_message(const std::string& topic,
                              FEConsumer::MsgType& msg)
{
    char type;
    if(!msg.getChar(BroadcastFieldId::MESSAGE_TYPE, type)) {
        std::cout << "Message missing message type: " << msg << std::endl;
        return;
    }
    using namespace boost::chrono;
    auto now = system_clock::now();
    auto now_ns = duration_cast<nanoseconds>(now.time_since_epoch());
    auto now_ulong = static_cast<unsigned long>(now_ns.count());
    if(m_iterate_messages) {
        iterate_message(msg);
    }
    if(m_print_messages) {
        print_message(topic, now_ulong, msg);
    }
    // Update latency if possible
    // But only for message types that matter
    int64_t msg_gen;
    bool check_latency = (type == BroadcastMessageType::TRADE ||
                          type == BroadcastMessageType::QUOTE);
    if(check_latency && msg.getInt64(BroadcastFieldId::MSG_GENESIS, msg_gen)) {
        unsigned long lat = 0;
        if(msg_gen < now_ulong) {
            lat = (now_ulong - msg_gen) / 1000000;  // Convert nanos to millis
        }
        if(lat > m_current_latency) {
            m_current_latency = lat;
            if(lat > m_slow_threshold) {
                std::unique_lock<std::mutex> guard(m_message_lock);
                m_slow_message = msg.clone();
                m_slow_time = now_ulong;
                m_slow_topic = topic;
            }
        }
    }
    // Check for gaps if possible
    int seq = -1;
    if(msg.getInt(BroadcastFieldId::FDS_SEQ_NUM, seq)) {
        auto it = m_sequence_map.find(topic);
        if(it != m_sequence_map.end()) {
            if(seq > it->second + 1) {
                if(m_print_gaps) {
                    std::cout << "Gap for " << topic << ", latest seq = " << seq
                              << ", previous seq = " << it->second << std::endl;
                }
                m_gaps_count += seq - (it->second + 1);
            }
        }
        m_sequence_map[topic] = seq;
    }
    // Add to message and bytes count
    m_msg_count++;
    m_bytes_count += msg.getSize();
}

void BroadcastCLI::iterate_message(const FEConsumer::MsgType& msg)
{
    for(auto it = msg.begin(), end = msg.end(); it != end; ++it) {
        typedef FactSet::Datafeed::BroadcastFieldType Types;
        switch(it->second.GetType()) {
            case Types::INT: {
                auto value = static_cast<int32_t>(it->second);
                if(value > 0)
                    ++G_int_count;
                break;
            }
            case Types::INT64: {
                auto value = static_cast<int64_t>(it->second);
                if(value > 0)
                    ++G_int64_count;
                break;
            }
            case Types::FLOAT: {
                auto value = static_cast<float>(it->second);
                if(value > 0)
                    ++G_double_count;
                break;
            }
            case Types::DOUBLE: {
                auto value = static_cast<double>(it->second);
                if(value > 0)
                    ++G_double_count;
                break;
            }
            case Types::STRING: {
                auto value = static_cast<std::string>(it->second);
                if(!value.empty())
                    ++G_string_count;
                break;
            }
            case Types::EPOCH_NSEC: {
                auto value = static_cast<MD::MD_EpochNS>(it->second);
                if(value.get() > 0)
                    ++G_ns_count;
                break;
            }
            case Types::CHAR: {
                auto value = static_cast<char>(it->second);
                if(value > '0')
                    ++G_char_count;
                break;
            }
            default:
                break;
        }
    }
}

void BroadcastCLI::print_message(const std::string& topic,
                                 long time_received,
                                 const FEConsumer::MsgType& msg)
{
    std::ostringstream printed_msg;
    printed_msg << std::setw(22) << std::left << "Topic"
                << " = " << topic << std::endl;
    printed_msg << std::setw(22) << std::left << "Time Received"
                << " = " << time_received << std::endl;
    printed_msg << msg << std::endl;
    std::cout << printed_msg.str();
}

void BroadcastCLI::print_suggest_help()
{
    std::cout << "Type help for usage instructions" << std::endl;
}

void BroadcastCLI::print_consumer_null()
{
    std::cout << "No consumer exists, use create command first" << std::endl;
}

void BroadcastCLI::print_too_few_args()
{
    std::cout << "Too few arguments to command" << std::endl;
}

void BroadcastCLI::print_stats()
{
    auto lat = m_current_latency.load();
    std::cout << "tmsgs: [" << std::setw(9) << m_msg_count << "]"
              << " msg/sec: [" << std::setw(6) << m_interval_msgs << "]"
              << " tgaps: [" << std::setw(3) << m_gaps_count << "]"
              << " gaps/sec: [" << std::setw(2) << m_interval_gaps << "]"
              << " Mb/sec: [" << std::setw(3)
              << (m_interval_bytes * 8) / 1000000 << "]"
              << " avg Mb/sec: [" << std::setw(3)
              << (m_bytes_count * 8) / 1000000 / m_updates << "]"
              << " lat: [" << std::setw(3) << lat << "]"
              << " hw: [" << std::setw(3) << m_highest_latency << "]"
              << std::endl;
    if(lat > m_slow_threshold) {
        std::unique_lock<std::mutex> guard(m_message_lock);
        print_message(m_slow_topic, m_slow_time, *m_slow_message);
    }
}

void BroadcastCLI::update_stats(const boost::system::error_code& e)
{
    // Reset timer
    m_stats_timer.expires_from_now(boost::posix_time::seconds(1));
    m_stats_timer.async_wait(
        std::bind(&BroadcastCLI::update_stats, this, std::placeholders::_1));
    // Check error
    if(m_print_stats && e) {
        std::cout << "Timer error: " << e << std::endl;
        return;
    }
    // Calculate for interval and print
    m_interval_msgs = m_msg_count - m_last_msg_count;
    m_last_msg_count = m_msg_count.load();
    m_interval_gaps = m_gaps_count - m_last_gaps_count;
    m_last_gaps_count = m_gaps_count.load();
    m_interval_bytes = m_bytes_count - m_last_bytes_count;
    m_last_bytes_count = m_bytes_count.load();
    if(m_current_latency > m_highest_latency) {
        m_highest_latency = m_current_latency.load();
    }
    if(m_print_stats) {
        print_stats();
    }
    m_current_latency = 0;
    if(m_interval_msgs > 0) {
        m_updates++;
    }
}

void BroadcastCLI::on_event(const Event& e, void* closure)
{
    std::ostringstream err_msg;
    if(e.error() == Error::NoError) {
        err_msg << "No Error";
    } else {
        err_msg << "Error " << e.error().code() << ": "
                << e.error().description();
    }
    std::cout << "Received event of type " << Event::to_string(e.type()) << " ("
              << err_msg.str() << ") " << e.all_values() << std::endl;
}

void BroadcastCLI::on_logging_message(const LogMessage* messages, size_t count)
{
    for(size_t i = 0; i < count; i++) {
        const LogMessage& msg = messages[i];
        std::ostringstream oss;
        oss << "Filename: " << msg.get_filename() << std::endl;
        oss << "Text: " << msg.get_text() << std::endl;
        oss << "Thread ID: " << msg.get_thread_id() << std::endl;
        oss << "Time: " << msg.get_time() << std::endl;
        oss << "Level: " << msg.get_level() << std::endl;
        oss << "Line: " << msg.get_line_number() << std::endl;
        oss << "Microseconds: " << msg.get_microseconds() << std::endl;
        std::cout << oss.str();
    }
}

BroadcastCLI* application;

void sigint_handler(int s)
{
    if(application->sigint_handler()) {
        delete application;
        exit(0);
    }
}

int main(int argc, char** argv)
{
    application = new BroadcastCLI();

    struct sigaction sigIntHandler;

    sigIntHandler.sa_handler = sigint_handler;
    sigemptyset(&sigIntHandler.sa_mask);
    sigIntHandler.sa_flags = 0;

    sigaction(SIGINT, &sigIntHandler, NULL);

    bool console_logging;
    bool print_gaps;
    bool print_messages;
    bool iterate_messages;
    unsigned long slow_threshold;
    boost::program_options::options_description desc("options");
    std::vector<std::string> inputs;
    std::vector<std::string> dispatch_prefixes;
    desc.add_options()("help,h", "Display help")(
        "console-logging,c",
        boost::program_options::bool_switch(&console_logging)
            ->default_value(false),
        "Print logging messages to console instead of to file")(
        "print-gaps,g",
        boost::program_options::bool_switch(&print_gaps)->default_value(false),
        "Output a message describing each gap detected")(
        "print-messages,m",
        boost::program_options::bool_switch(&print_messages)
            ->default_value(false),
        "Output all fields of all messages received")(
        "print-slowest,s",
        boost::program_options::value<unsigned long>(&slow_threshold)
            ->default_value(ULONG_MAX),
        "When printing stats, the slowest message with latency over the given "
        "threshold in milliseconds will be printed.")(
        "iterate-messages,i",
        boost::program_options::bool_switch(&iterate_messages)
            ->default_value(false),
        "Iterate fields of each message to simulate work in the callback")(
        "auto,a",
        boost::program_options::value<std::vector<std::string>>(&inputs),
        "Command to run on startup. May appear multiple times and commands "
        "are executed in the order presented.")(
        "dispatch-prefix,d",
        boost::program_options::value<std::vector<std::string>>(
            &dispatch_prefixes),
        "A new thread will be created to dispatch topics with the given "
        "prefix. Multiple prefixes may be given and a new thread will be "
        "created for each of them. If no prefixes are given, a single thread "
        "will be created that will handle every topic.");
    boost::program_options::positional_options_description pos_opt;
    pos_opt.add("auto", 1);
    pos_opt.add("dispatch-prefix", 1);
    pos_opt.add("print-slowest", 1);

    boost::program_options::variables_map vm;
    try {
        boost::program_options::store(
            boost::program_options::command_line_parser(argc, argv)
                .options(desc)
                .positional(pos_opt)
                .run(),
            vm);
        boost::program_options::notify(vm);
    } catch(std::exception& e) {
        std::cerr << e.what() << std::endl;
        std::cerr << desc << std::endl;
        return EXIT_FAILURE;
    }

    if(vm.count("help")) {
        std::cout << desc << std::endl;
        return EXIT_SUCCESS;
    }

    application->start(console_logging,
                       print_gaps,
                       print_messages,
                       iterate_messages,
                       inputs,
                       dispatch_prefixes,
                       slow_threshold);
}
